#!/usr/bin/perl

BEGIN
{
    ($dirprg, $main::SCRIPT) = $0 =~ /(.+)\/(.+)/;
    unshift(@INC, "$dirprg/../lib", "$dirprg/../../lib/");

}

use strict;
use ParamParser;
use Lucene;
use iANT::Cfg;
use XML::Twig;
use iANT::Common;
require "iANT/Utils.pl";
use MIME::Base64;
use File::Temp;
use Cwd 'abs_path';
use Digest::MD5 qw(md5 md5_hex md5_base64);


my $O_PARAM;

MAIN:
{


#open (PRFLUCENE, '>>/tmp/profile.edition.rhime.lucene');
#my $time0 = time;
#print PRFLUCENE "Start:$time0\n";
#my $new_time = '',
#my $len = '';


    my $create_index = 0;

    $O_PARAM = New ParamParser(
                               'GETOPTLONG',
                               (
                                "parser_cfg=s", "store_dir=s",
                                "input=s",      "escape_special_characters", "supplementary_fields=s",
                                "maxdepth=i",   "mindepth=i",
                                "ctime=i",      "verbose",
								"no_update"
                                 )
                                 );
    $O_PARAM->SetUsage(my $usage = sub {&Usage();});

    $O_PARAM->AssertFileExists('parser_cfg');

    my $parser_cfg = $O_PARAM->Get('parser_cfg');

    $O_PARAM->AssertDefined('store_dir');
    $O_PARAM->AssertDefined('input');

    my $o_parser     = New ParamParser($parser_cfg);
    my $analyzer_mod = $o_parser->Get('analyzer');
    my $analyzer     = new $analyzer_mod;

    $O_PARAM->SetBehaviour('use_substitution_table');
    $O_PARAM = &iANT::Cfg::Load($O_PARAM);

    my $unix_find = $O_PARAM->Get('unix_find');
    my $unix_gzip = $O_PARAM->Get('unix_gzip');
    my $unix_cat  = $O_PARAM->Get('unix_cat');

    #est-ce que ce sera forcement un project ? non
    #parametre lucene_dir

    my $store_dir = $O_PARAM->Get('store_dir');

    if (-d $store_dir)
    {
    	opendir(ACCESSDIR, $store_dir);
	my @a_store_files = readdir(ACCESSDIR);
	closedir(ACCESSDIR);
        #Even if the directory is empty, . and .. exists, so the array has a minimal length of 2
	$create_index = 1 if (@a_store_files == 2);
	if (!$create_index)
	{
		my $nb_cfs = grep (/\.cfs/, @a_store_files);
		if ($nb_cfs == 0)
		{
			print STDERR "ERROR: There are other files than .cfs in the directory\n" if ($O_PARAM->IsDefined('verbose'));
			die ("No cfs file in this directory");
		}
	}
    }
    
    $create_index = 1 if (!-d $store_dir);

    print STDERR "create_index $create_index $store_dir\n" if ($O_PARAM->IsDefined('verbose'));
    mkdir $store_dir if ($create_index);

    my $store = Lucene::Store::FSDirectory->getDirectory($store_dir, $create_index);

    #$doc serait une entree ?
    my @a_files = ();
    if (-d $O_PARAM->Get('input'))
    {
        my $data_dir  = $O_PARAM->Get('input');
        my $extension = $o_parser->Get('suffix');
        print STDERR "EXTENSION: $extension\n\n" if ($O_PARAM->IsDefined('verbose'));

        my $maxdepth = '';
        my $mindepth = '';
        my $ctime    = '';

        $maxdepth = '-maxdepth ' . $O_PARAM->Get('maxdepth') if ($O_PARAM->IsDefined('maxdepth'));
        $mindepth = '-mindepth ' . $O_PARAM->Get('mindepth') if ($O_PARAM->IsDefined('mindepth'));
        $ctime    = '-ctime ' . $O_PARAM->Get('ctime')       if ($O_PARAM->IsDefined('ctime'));

        my $cmd = "$unix_find $data_dir $maxdepth $mindepth -type f -name \\*.$extension";
        print STDERR "CMD: $cmd\n\n" if ($O_PARAM->IsDefined('verbose'));
        @a_files = `$cmd`;
    }
    elsif (-e $O_PARAM->Get('input'))
    {
        if ($O_PARAM->IsDefined('verbose'))
        {
            print STDERR 'WARNING: "--maxdepth" parameter is only available when --input is a directory'
              if ($O_PARAM->IsDefined('maxdepth'));
            print STDERR 'WARNING: "--mindepth" parameter is only available when --input is a directory'
              if ($O_PARAM->IsDefined('mindepth'));
            print STDERR 'WARNING: "--ctime" parameter is only available when --input is a directory'
              if ($O_PARAM->IsDefined('ctime'));
        }

        my $ra_files = &iANT::Common::GetFileNamesFromFof($O_PARAM->Get('input'));

        if ($ra_files)
        {
            @a_files = @{$ra_files};
            print STDERR "INFO: Dealing with a file of files\n" if ($O_PARAM->IsDefined('verbose'));
            1;
        }
        else
        {
            print STDERR "INFO: Dealing with an xml file\n" if ($O_PARAM->IsDefined('verbose'));

            @a_files = ($O_PARAM->Get('input'));
        }

    }

    chomp @a_files;

    my ($ra_fields_to_parse, $rh_docid) = &GetLuceneFields4Xml($o_parser);

    if ((!$create_index) && (!$O_PARAM->IsDefined('no_update')))
    {
        my $search_query     = '';
        my @a_search_queries = ();

        #utiliser un @queries contenant au max 32 clauses. Limitation Lucene 1.4 a priori ...
        #%32
        my $count            = 0;
        my %h_search_queries = ();
        foreach my $to_add (@a_files)
        {

            if (-B $to_add)
            {
                print STDERR "ERROR: $to_add is a binary file !\n";
                next;
            }
            my $o_twig = new XML::Twig;
            $o_twig->parsefile($to_add);
            my $root = $o_twig->root;
            my ($docid_elt) = $root->get_xpath($rh_docid->{xpath});
            my $docid_value;
	        if ($rh_docid->{xpath} eq 'abs_path')
            {
                #SC 20100806
                #Attention  : si on veut se servir du abs_path (ou son md5) comme docid, il faut a tout prix utiliser l'indexer Keyword
                #sinon il n'arrive pas a supprimer un doc deja existant
                #je ne comprends pas pourquoi ...
                #si on veut stocker le md5 du path du fichier fichier, il faut dans le parser ecrire field:mon_alias=Keyword.abs_path@md5

        		$docid_value = abs_path($to_add);
                if ((defined $rh_docid->{att}) && ($rh_docid->{att} eq 'md5'))
                {
                    $docid_value = md5_hex($docid_value);
                }
            }
            elsif ((!defined $rh_docid->{att}) || ($rh_docid->{att} eq ''))
            {
                $docid_value = $docid_elt->text();
                $docid_value = &EscapeLuceneSpecialCharaters($docid_value)
                  if ($O_PARAM->IsDefined('escape_special_characters'));
            }
            else
            {
                $docid_value = $docid_elt->{att}->{$rh_docid->{att}};
                $docid_value = &EscapeLuceneSpecialCharaters($docid_value)
                  if ($O_PARAM->IsDefined('escape_special_characters'));
            }
            if ((defined $docid_value) && ($docid_value ne ''))
            {
		        my $reader;
                eval {$reader = Lucene::Index::IndexReader->open($store)};

                if ($@)
                {
                    print STDERR $@;
                }
                else
                {

                    $reader = Lucene::Index::IndexReader->open($store);

                    #if indexer = Text ---> lc($value) mode
		    #le terme Lucene est mis en minuscule quand l'indexer est du texte
		    #il n'y a que l'indexer Keyword qui ne change pas la casse
		    
                    $docid_value = lc($docid_value) if ($rh_docid->{mode} !~ /keyword/i);


        		    print STDOUT "INFO: Delete entries with term " . $rh_docid->{alias} . " '$docid_value'\n" if ($O_PARAM->IsDefined('verbose'));
#$time0 = time;
                    my $term = new Lucene::Index::Term($rh_docid->{alias}, $docid_value);

#$new_time = time;
#$len = $new_time - $time0;
#print PRFLUCENE " deleteDocuments: $len\n";
                    $reader->deleteDocuments($term) or print "WARNING: Can not delete entry " . $rh_docid->{alias} . ' >' . $docid_value . "<\n";

                    $reader->close();
                    undef $reader;

                }
            }
        }

        my $reader;

    }
#$time0 = time;

    my $writer = new Lucene::Index::IndexWriter($store, $analyzer, $create_index);
#$new_time = time;
#$len = $new_time - $time0;
#print PRFLUCENE " Lucene::Index::IndexWriter: $len\n";

#$time0 = time;

#Pas utile celui-ci
#    $writer->optimize();

#$new_time = time;
#$len = $new_time - $time0;
#print PRFLUCENE " optimize: $len\n";

    foreach my $to_add (@a_files)
    {
        my $doc_id_value;

	if (-B $to_add)
        {
            print STDERR "ERROR: $to_add is a binary file !\n";
            next;
        }

        print STDERR "INFO:analyse $to_add\n" if ($O_PARAM->IsDefined('verbose'));

        my $o_twig = new XML::Twig;
        $o_twig->parsefile($to_add);
        my $root = $o_twig->root;

        my $doc = new Lucene::Document;
        print STDERR "INFO:add $to_add\n" if ($O_PARAM->IsDefined('verbose'));


	if ($O_PARAM->IsDefined('supplementary_fields') && ($O_PARAM->Get('supplementary_fields') ne ''))
	{
		my @a_supplementary_fields = split (',',$O_PARAM->Get('supplementary_fields'));
		foreach my $supplementary_field (@a_supplementary_fields)
		{
			my ($alias,$value) = ($supplementary_field =~/(\S+):(.+)/);
			if (defined $alias && defined $value)
			{
				$doc->add(Lucene::Document::Field->Text("$alias", "$value"));
			} 
		}
	}

        #Utiliser @a_fields_to_parse
        foreach my $rh_field_to_parse (@{$ra_fields_to_parse})
        {
            my $xpath = $rh_field_to_parse->{xpath};
            my $att   = $rh_field_to_parse->{att};
            my $alias = $rh_field_to_parse->{alias};
            my $mode  = $rh_field_to_parse->{mode};
            my @a_elt = $root->get_xpath($xpath);
            my $value;

            if ($xpath eq 'file')
            {

                #gzip + b64 => compression d'un facteur 2 a 2.5

                my $tmp_dir = $O_PARAM->Get('tmp_dir');
                my ($tmp_fh, $tmp_filename) = File::Temp::tempfile(DIR => $tmp_dir);
                my $gz_xml = `$unix_gzip -c $to_add`;
                chomp $gz_xml;
                unlink($tmp_filename);
                my $value = MIME::Base64::encode($gz_xml);
                $doc->add(Lucene::Document::Field->$mode("$alias", "$value"));
            }
            elsif ($xpath eq 'abs_path')
            {
		        my $value = abs_path($to_add);
                #SC 20100806
                #Attention  : si on veut se servir du abs_path (ou son md5) comme docid, il faut a tout prix utiliser l'indexer Keyword
                #sinon il n'arrive pas a supprimer un doc deja existant
                #je ne comprends pas pourquoi ...
                #si on veut stocker le md5 du path du fichier fichier, il faut dans le parser ecrire field:mon_alias=Keyword.abs_path@md5
                if ((defined $att) && ($att eq 'md5'))
                {
                    $value = md5_hex($value);
                }
                $doc_id_value = $value if ($alias eq $rh_docid->{alias});
		        $doc->add(Lucene::Document::Field->$mode("$alias", "$value"));
            }
            else
            {
                foreach my $elt (@a_elt)
                {

                    if ((!defined $att) || ($att eq ''))
                    {
                        $value .= ' // ' . $elt->text() if (defined $elt);
                        $value = $elt->text() if ((defined $elt) && ($#a_elt == 0));
                    }
                    else
                    {
                        my $sep = ' ';
                        $sep .= $elt->{att}->{id} . ':' if (defined $elt->{att}->{id} && ($att ne 'id'));
                        $value .= $sep . $elt->{att}->{$att} if (defined $elt);

                        $value = $elt->{att}->{$att} if ((defined $elt) && ($#a_elt == 0));
                    }
                }
		$doc_id_value = $value if ($alias eq $rh_docid->{alias});
                if ($value ne '')
                {
                    $doc->add(Lucene::Document::Field->$mode("$alias", $value));
                }
            }
        }

        # field gets just stored
        #tout le document XML ?
	
        if (defined $doc_id_value && ($doc_id_value ne ''))
        {
#$time0 = time;
            $writer->addDocument($doc);
#$new_time = time;
#$len = $new_time - $time0;
#print PRFLUCENE " addDocument: $len\n";

        }
        else
        {
            print STDERR "WARNING a document in $to_add has not been added because is docid is empty\n";
        }   

    }

#$time0 = time;

    $writer->optimize();

#$new_time = time;
#$len = $new_time - $time0;
#print PRFLUCENE " optimize: $len\n";

    $writer->close();
    undef $writer;
    
#close PRFLUCENE;

}

sub Usage
{

    print STDOUT <<END;
   
Usage:

   $0 --parser_cfg=<file> --store_dir=<directory> --input=<file|directory> [--maxdepth=i] [--mindepth=i] [--escape_special_characters]

Description

    Add/Update xml files to a Lucene index
   
 Mandatory
    --parser_cfg:   parser configuration file (contains Analyzer name, fields to parse, ...)
    --store_dir :   directory where Lucene indeces are stored. If you want to create a new index, the store repository must not exist, lucene is going to create it.
    --input     :   file : add file to index
                    directory: scan directory looking for files with extension defined in 
                    parser_cfg file (parameter:suffix)
		    
  Optional
     --supplementary_fields=<list of field:value> a list (comma separated) of field:value to add in the Lucene db for each document
     --escape_special_characters : to escape special characters in Lucene query
			           list: \ + - ! ( ) { } [ ] ^ " ~ * ? :

     --maxdepth	:   [unix find option] Descend at most "i" levels of directories below the command line arguments (--input is a directory)
     --mindepth	:   [unix find option] Do  not apply any tests or actions at levels less than "i" (--input is a directory)
     --ctime	:   [unix find option] File’s status was last changed i*24 hours ago (--input is a directory)
     --no_update:   Doesn't check if a entry belongs to index. Add it anyway.
                                

END
}
